"""An implementation of an unconstrained n-dimensional continuous space."""

from typing import Final

import numpy

from moptipy.spaces.vectorspace import VectorSpace
from moptipy.utils.bounds import FloatBounds
from moptipy.utils.logger import KeyValueLogSection


class BoundedVectorSpace(VectorSpace, FloatBounds):
    """
    A bounded vector space whose elements are a one-dimensional numpy array.

    Such spaces are useful for continuous optimization.
    """

    def __init__(self, dimension: int,
                 min_value: float = -1.0, max_value: float = 1.0,
                 dtype: numpy.dtype = numpy.dtype(numpy.float64)) -> None:
        """
        Create the vector-based search space.

        :param dimension: The dimension of the search space,
            i.e., the number of decision variables.
        :param min_value: the minimum permitted value
        :param max_value: the maximum permitted value
        :param dtype: The basic data type of the vector space,
            i.e., the type of the decision variables
        """
        VectorSpace.__init__(self, dimension, dtype)
        FloatBounds.__init__(self, min_value, max_value)

    def validate(self, x: numpy.ndarray) -> None:
        """
        Validate a vector.

        :param x: the real vector
        :raises TypeError: if the vector is not an element of this space.
        :raises ValueError: if the shape of the vector is wrong or any of its
            element is not finite.
        """
        super().validate(x)
        xmin: Final[float] = x.min()
        xmax: Final[float] = x.max()
        if not (self.min_value <= xmin <= xmax <= self.max_value):
            raise ValueError(
                f"permitted range of elements is [{self.min_value}"
                f", {self.max_value}] but found elements in range "
                f"[{xmin}, {xmax}].")

    def __str__(self) -> str:
        """
        Get the name of this space.

        :return: "vector" + dimension + dtype.char

        >>> print(BoundedVectorSpace(3, -1.0, 1.0))
        vector3d_m1_1

        >>> print(BoundedVectorSpace(3, -1.0, 1.6))
        vector3d_m1_1d6
        """
        return f"{VectorSpace.__str__(self)}_{FloatBounds.__str__(self)}"

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        """
        Log the parameters of this space to the given logger.

        :param logger: the logger for the parameters
        """
        VectorSpace.log_parameters_to(self, logger)
        FloatBounds.log_parameters_to(self, logger)  # type: ignore



"""
A set of numerical optimization algorithms from SciPy.

The function :func:`scipy.optimize.minimize` provides a set of very
efficient numerical/continuous optimization methods. Here we wrap a set of
them into our `moptipy` :class:`~moptipy.api.process.Process` API. All
algorithms provided in this module are imported and wrapped from SciPy
(https://scipy.org).

By using the :func:`~moptipy.api.subprocesses.without_should_terminate`
tool, we can enforce the termination criteria set via the
:class:`~moptipy.api.execution.Execution` builder on external algorithms
while piping all their function evaluations through the
:meth:`~moptipy.api.process.Process.evaluate` routine of the optimization
:meth:`~moptipy.api.process.Process`. This way, we can make these external
algorithms usable within `moptipy` in a transparent manner.
"""
from typing import Callable, Optional, cast, Any, Final, List, Tuple

from numpy import full, inf, ndarray, clip
from scipy.optimize import Bounds  # type: ignore
# noinspection PyProtectedMember
from scipy.optimize._differentialevolution import (  # type: ignore
    differential_evolution)  # type: ignore
# noinspection PyProtectedMember
from scipy.optimize._optimize import _minimize_bfgs  # type: ignore
# noinspection PyProtectedMember
from scipy.optimize._optimize import _minimize_cg  # type: ignore
# noinspection PyProtectedMember
from scipy.optimize._optimize import _minimize_neldermead  # type: ignore
# noinspection PyProtectedMember
from scipy.optimize._optimize import _minimize_powell  # type: ignore
# noinspection PyProtectedMember
from scipy.optimize._slsqp_py import _minimize_slsqp  # type: ignore
# noinspection PyProtectedMember
from scipy.optimize._tnc import _minimize_tnc  # type: ignore

from moptipy.api.algorithm import Algorithm, Algorithm0
from moptipy.api.operators import Op0
from moptipy.api.process import Process
from moptipy.api.subprocesses import without_should_terminate
from moptipy.operators.vectors.op0_uniform import Op0Uniform
from moptipy.utils.bounds import OptionalFloatBounds, FloatBounds
from moptipy.utils.logger import KeyValueLogSection
from moptipy.utils.nputils import DEFAULT_FLOAT
from moptipy.utils.types import type_error


# noinspection PyProtectedMember


class SciPyAlgorithmWrapper(Algorithm0, OptionalFloatBounds):
    """
    A wrapper for the Sci-Py API.

    An instance of this class may be re-used, but it must only be used for
    problems of the same dimension.
    """

    def __init__(self, name: str, op0: Op0,
                 min_value: Optional[float] = None,
                 max_value: Optional[float] = None) -> None:
        """
        Create the algorithm importer from scipy.

        :param name: the name of the algorithm
        :param op0: the nullary search operator
        :param min_value: the minimum x-value
        :param max_value: the maximum x-value
        """
        Algorithm0.__init__(self, name, op0)
        # load bounds from nullary operator, if possible and not overriden
        if (min_value is None) and (max_value is None) \
                and isinstance(op0, Op0Uniform):
            min_value = op0.min_value
            max_value = op0.max_value
        OptionalFloatBounds.__init__(self, min_value, max_value)
        #: the bounds to be used for the internal function call
        self.__bounds_cache: Optional[Bounds] = None
        #: the cache for starting points
        self.__x0_cache: Optional[ndarray] = None

    def _call(self, func: Callable, x0, max_fes: int,
              bounds: Optional[Bounds]) -> None:
        """
        Invoke the SciPi Algorithm.

        This function will be overwritten to call the SciPi Algorithm.

        :param func: the function to minimize
        :param x0: the starting point
        :param max_fes: the maximum FEs
        :param bounds: the bounds
        """

    def __run(self, pp: Process) -> None:
        """
        Execute the algorithm.

        :param pp: the process
        """
        x0 = self.__x0_cache
        if x0 is None:
            self.__x0_cache = x0 = pp.create()  # create the solution record
            x0_dim = len(x0)  # the dimension of the solution record
            no_bounds: bool = False  # True if no bounds needed
            mi: float = -inf  # the minimum
            ma: float = inf  # the maximum
            if self.min_value is None:
                if self.max_value is None:
                    no_bounds = True  # no bounds needed if none provided
            else:
                mi = self.min_value  # remember minimum (otherwise, mi=-inf)
            if self.max_value is not None:
                ma = self.max_value  # remember maximum (otherwise ma=inf)
            # now create bounds record
            self.__bounds_cache = bounds = None if no_bounds else Bounds(
                full(x0_dim, mi, DEFAULT_FLOAT),  # the lower bound vector
                full(x0_dim, ma, DEFAULT_FLOAT))  # the upper bound vector

        else:  # ok, we have cached bounds (or cached None)
            bounds = self.__bounds_cache

        if bounds is None:
            __func = pp.evaluate
        else:
            def __func(x: ndarray, ff=cast(Callable[[ndarray], Any],
                                           pp.evaluate),
                       lb=self.min_value, ub=self.max_value):
                clip(x, lb, ub, x)
                return ff(x)

        self.op0.op0(pp.get_random(), x0)  # create first solution

        mf = pp.get_max_fes()  # get the number of available FEs
        if mf is not None:  # if an FE limit is specified, then ...
            mf -= pp.get_consumed_fes()  # ... subtract the consumed FEs
        else:  # otherwise set a huge, unattainable limit
            mf = 4_611_686_018_427_387_904  # 2 ** 62
        self._call(__func, x0, mf, bounds)  # invoke the algorithm

    def solve(self, process: Process) -> None:
        """
        Apply the algorithm from SciPy to an optimization problem.

        Basically, this wraps a specific configuration of
        :func:`scipy.optimize.minimize` into our process API and
        invokes it.

        :param process: the black-box process object
        """
        # invoke the SciPy algorithm implementation
        without_should_terminate(
            cast(Callable[[Process], Any], self.__run), process)

    def log_parameters_to(self, logger: KeyValueLogSection):
        """
        Log the parameters of the algorithm to a logger.

        :param logger: the logger for the parameters
        """
        Algorithm0.log_parameters_to(self, logger)  # log algorithm/operator
        OptionalFloatBounds.log_parameters_to(self, logger)  # log bounds


def _call_powell(func: Callable, x0, max_fes: int,
                 bounds: Optional[Bounds]) -> None:
    _minimize_powell(func, x0, bounds=bounds, xtol=0.0, ftol=0.0,
                     maxiter=max_fes, maxfev=max_fes)


class Powell(SciPyAlgorithmWrapper):
    """
    Powell's Algorithm.

    The function :func:`scipy.optimize.minimize` with parameter
    "Powell" can perform unconstrained continuous optimization
    (potentially with boundary constraints)

    1. Michael James David Powell. An Efficient Method for Finding the Minimum
       of a Function of Several Variables without Calculating Derivatives. The
       Computer Journal. 7(2):155-162. 1964.
       doi:https://doi.org/10.1093/comjnl/7.2.155
    """

    def __init__(self, op0: Op0,
                 min_value: Optional[float] = None,
                 max_value: Optional[float] = None) -> None:
        """
        Create Powell's Algorithm from SciPy.

        :param op0: the nullary search operator
        :param min_value: the minimum x-value
        :param max_value: the maximum x-value
        """
        super().__init__("spPowell", op0, min_value, max_value)
        self._call = _call_powell  # type: ignore


def _call_nelder_mead(func: Callable, x0, max_fes: int,
                      bounds: Optional[Bounds]) -> None:
    _minimize_neldermead(func, x0, bounds=bounds, xatol=0.0, fatol=0.0,
                         maxiter=max_fes, maxfev=max_fes)


class NelderMead(SciPyAlgorithmWrapper):
    """
    The Downhill Simplex aka. the Nelder-Mead Algorithm.

    The function :func:`scipy.optimize.minimize` with parameter
    "Nelder-Mead" can perform unconstrained continuous optimization
    (potentially with boundary constraints) by using the Downhill Simplex
    algorithm a.k.a., the Nelder-Mead algorithm. Here we wrap it into our API.

    Scipy provides the following reference:

    1. Fuchang Gao and Lixing Han. Implementing the Nelder-Mead Simplex
       Algorithm with Adaptive Parameters. *Computational Optimization and
       Applications*. 51(1):259–277. January 2012.
       doi:https://doi.org/10.1007/s10589-010-932
    2. J. A. Nelder and R. Mead. A Simplex Method for Function Minimization.
       *The Computer Journal*. 7(4):308-313. January 1965. Oxford University
       Press (OUP). doi:http://dx.doi.org/10.1093/COMJNL/7.4.308
       https://people.duke.edu/~hpgavin/cee201/Nelder+Mead-\
ComputerJournal-1965.pdf
    3. M. H. Wright. Direct Search Methods: Once Scorned, Now Respectable.
       In D.F. Griffiths and G.A. Watson (Eds.) *Proceedings of the 1995
       Dundee Biennial Conference in Numerical Analysis*. Harlow, UK:
       Addison Wesley Longman, pp. 191-208.
    4. Jorge Nocedal and Stephen J. Wright. *Numerical Optimization*. In
       Springer Series in Operations Research and Financial Engineering.
       New York, NY, USA: Springer. 2006. Second Edition.
       ISBN: 978-0-387-30303-1. Chapter 9.5, Page 238.
       doi:https://doi.org/10.1007/978-0-387-40065-5.
    """

    def __init__(self, op0: Op0,
                 min_value: Optional[float] = None,
                 max_value: Optional[float] = None) -> None:
        """
        Create the Nelder-Mead Algorithm from SciPy.

        :param op0: the nullary search operator
        :param min_value: the minimum x-value
        :param max_value: the maximum x-value
        """
        super().__init__("spNelderMead", op0, min_value, max_value)
        self._call = _call_nelder_mead  # type: ignore


def _call_bgfs(func: Callable, x0, max_fes: int, _) -> None:
    _minimize_bfgs(func, x0, gtol=0.0, maxiter=max_fes)


class BGFS(SciPyAlgorithmWrapper):
    """
    The wrapper for the BGFS algorithm in SciPy.

    This is the quasi-Newton method by C. G. Broyden, Roger Fletcher,
    D. Goldfarb, and David F. Shanno (BFGS).

    1. Jorge Nocedal and Stephen J. Wright. *Numerical Optimization*. In
       Springer Series in Operations Research and Financial Engineering.
       New York, NY, USA: Springer. 2006. Second Edition.
       ISBN: 978-0-387-30303-1. Chapter 6, Page 136.
       doi:https://doi.org/10.1007/978-0-387-40065-5.
    2. Roger Fletcher. *Practical Methods of Optimization* (2nd ed.),
       New York: John Wiley & Sons. 1987. ISBN 978-0-471-91547-8.
    3. C. G. Broyden. The convergence of a class of double-rank minimization
       algorithms. *Journal of the Institute of Mathematics and Its
       Applications*. 6(1):76-90. March 1970.
       doi:http://dx.doi.org/10.1093/imamat/6.1.76
    """

    def __init__(self, op0: Op0,
                 min_value: Optional[float] = None,
                 max_value: Optional[float] = None) -> None:
        """
        Create the BGFS Algorithm from SciPy.

        :param op0: the nullary search operator
        :param min_value: the minimum x-value
        :param max_value: the maximum x-value
        """
        super().__init__("spBgfs", op0, min_value, max_value)
        self._call = _call_bgfs  # type: ignore


def _call_cg(func: Callable, x0, max_fes: int, _) -> None:
    _minimize_cg(func, x0, gtol=0.0, maxiter=max_fes)


class CG(SciPyAlgorithmWrapper):
    """
    The wrapper for the Conjugate Gradient algorithm in SciPy.

    1. Jorge Nocedal and Stephen J. Wright. *Numerical Optimization*. In
       Springer Series in Operations Research and Financial Engineering.
       New York, NY, USA: Springer. 2006. Second Edition.
       ISBN: 978-0-387-30303-1. Chapter 5, Page 101.
    """

    def __init__(self, op0: Op0,
                 min_value: Optional[float] = None,
                 max_value: Optional[float] = None) -> None:
        """
        Create the CG Algorithm from SciPy.

        :param op0: the nullary search operator
        :param min_value: the minimum x-value
        :param max_value: the maximum x-value
        """
        super().__init__("spCg", op0, min_value, max_value)
        self._call = _call_cg  # type: ignore


def _call_slsqp(func: Callable, x0, max_fes: int, _) -> None:
    _minimize_slsqp(func, x0, ftol=0.0, maxiter=max_fes)


class SLSQP(SciPyAlgorithmWrapper):
    """
    The Sequential Least Squares Programming (SLSQP) algorithm in SciPy.

    1. Dieter Kraft. Algorithm 733: TOMP–Fortran modules for optimal control
       calculations. *ACM Transactions on Mathematical Software.*
       20(3):262-281. September 1994. https://doi.org/10.1145/192115.192124
    """

    def __init__(self, op0: Op0,
                 min_value: Optional[float] = None,
                 max_value: Optional[float] = None) -> None:
        """
        Create the SLSQP Algorithm from SciPy.

        :param op0: the nullary search operator
        :param min_value: the minimum x-value
        :param max_value: the maximum x-value
        """
        super().__init__("spSlsqp", op0, min_value, max_value)
        self._call = _call_slsqp  # type: ignore


def _call_tnc(func: Callable, x0, max_fes: int,
              bounds: Optional[Bounds]) -> None:
    if bounds is None:
        b = None
    else:
        b = [(bounds.lb[0], bounds.ub[0])] * len(x0)
    _minimize_tnc(func, x0, bounds=b, ftol=0.0, xtol=0.0, gtol=0.0,
                  maxiter=max_fes, maxfun=max_fes)


class TNC(SciPyAlgorithmWrapper):
    """
    The Truncated Newton Method from SciPy.

    1. Stephen G. Nash. Newton-Type Minimization via the Lanczos Method.
       *SIAM Journal on Numerical Analysis*. 21(4):770-783. August 1984.
       https://dx.doi.org/10.1137/0721052.
    2. Jorge Nocedal and Stephen J. Wright. *Numerical Optimization*. In
       Springer Series in Operations Research and Financial Engineering.
       New York, NY, USA: Springer. 2006. Second Edition.
       ISBN: 978-0-387-30303-1. doi:https://doi.org/10.1007/978-0-387-40065-5.
    """

    def __init__(self, op0: Op0,
                 min_value: Optional[float] = None,
                 max_value: Optional[float] = None) -> None:
        """
        Create the TNC Algorithm from SciPy.

        :param op0: the nullary search operator
        :param min_value: the minimum x-value
        :param max_value: the maximum x-value
        """
        super().__init__("spTnc", op0, min_value, max_value)
        self._call = _call_tnc  # type: ignore


class DE(Algorithm, FloatBounds):
    """
    The Differential Evolution Algorithm as implemented by SciPy.

    At this point, we do not expose the many parameters of the function
    :func:`scipy.optimize.differential_evolution`.
    We only use the default settings. This may change in future releases.

    1. Rainer Storn and Kenneth Price. Differential Evolution – A Simple and
       Efficient Heuristic for global Optimization over Continuous Spaces.
       *Journal of Global Optimization* 11(4):341–359. December 1997.
       doi:https://doi.org/10.1023/A:1008202821328.
       https://www.researchgate.net/publication/227242104
    """

    def __init__(self, dim: int,
                 min_value: float = -1e10,
                 max_value: float = 1e10) -> None:
        """
        Create the Differential Evolution Algorithm from SciPy.

        :param dim: the dimension in which the algorithm works
        :param min_value: the minimum x-value
        :param max_value: the maximum x-value
        """
        FloatBounds.__init__(self, min_value, max_value)
        if not isinstance(dim, int):
            raise type_error(dim, "dim", int)
        if (dim < 1) or (dim > 100_000):
            raise ValueError(
                f"dim must be in 1...100_000 but is {dim}.")
        #: the bounds
        self.__bounds: Final[List[Tuple[float, float]]] = \
            [(self.min_value, self.max_value)] * dim

    def __run(self, pp: Process):
        """
        Execute the algorithm.

        :param pp: the process
        """

        def __func(x: ndarray, ff=cast(Callable[[ndarray], Any],
                                       pp.evaluate),
                   lb=self.min_value, ub=self.max_value):
            clip(x, lb, ub, x)
            return ff(x)

        mf = pp.get_max_fes()  # get the number of available FEs
        if mf is not None:  # if an FE limit is specified, then ...
            mf -= pp.get_consumed_fes()  # ... subtract the consumed FEs
        else:  # otherwise set a huge, unattainable limit
            mf = 4_611_686_018_427_387_904  # 2 ** 62

        differential_evolution(
            __func, bounds=self.__bounds,
            maxiter=int(mf / len(self.__bounds)) + 1,
            tol=0.0, seed=pp.get_random(), atol=0.0)

    def solve(self, process: Process) -> None:
        """
        Apply the algorithm from SciPy to an optimization problem.

        Basically, this wraps a specific configuration of
        :func:`scipy.optimize.minimize` into our process API and
        invokes it.

        :param process: the black-box process object
        """
        # invoke the SciPy algorithm implementation
        without_should_terminate(
            cast(Callable[[Process], Any], self.__run), process)

    def log_parameters_to(self, logger: KeyValueLogSection):
        """
        Log the parameters of the algorithm to a logger.

        :param logger: the logger for the parameters
        """
        Algorithm.log_parameters_to(self, logger)  # log algorithm/operator
        FloatBounds.log_parameters_to(self, logger)  # log bounds

    def __str__(self):
        """
        Get the name of this algorithm.

        :returns: the name of this differential evolution algorithm
        """
        return "spDE"



"""A nullary operator filling a vector with uniformly distribute values."""

import numpy as np
from numpy.random import Generator

from moptipy.api.operators import Op0
from moptipy.utils.bounds import FloatBounds
from moptipy.utils.logger import KeyValueLogSection


class Op0Uniform(Op0, FloatBounds):
    """Fill a vector with uniformly distributed random values."""

    def __init__(self, min_value: float = -1.0,
                 max_value: float = 1.0) -> None:
        """
        Initialize the operator.

        :param min_value: the minimum value
        :param max_value: the maximum value
        """
        Op0.__init__(self)
        FloatBounds.__init__(self, min_value, max_value)

    def __str__(self) -> str:
        """
        Get the name of this operator.

        :returns: the name of this space
        """
        return "uniform"

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        """
        Log the parameters of this space to the given logger.

        :param logger: the logger for the parameters
        """
        Op0.log_parameters_to(self, logger)
        FloatBounds.log_parameters_to(self, logger)

    def op0(self, random: Generator, dest: np.ndarray) -> None:
        """
        Fill the string `dest` with random values.

        :param random: the random number generator
        :param dest: the vector to be filled. Afterwards it contains uniformly
            distributed random values
        """
        np.copyto(dest, random.uniform(
            self.min_value, self.max_value, len(dest)))



"""A set of bounds, i.e., a minimal and a maximal value."""

from math import isfinite, inf
from typing import Final, Optional

from moptipy.utils.logger import KeyValueLogSection
from moptipy.utils.strings import num_to_str_for_name
from moptipy.utils.types import type_error

#: the log key for the minimum value
KEY_MIN: Final[str] = "min"
#: the log key for the maximum value
KEY_MAX: Final[str] = "max"


class Bounds:
    """A set of bounds, i.e., a minimal and a maximal value."""

    def __init__(self, min_value, max_value) -> None:
        """
        Create the bounds object.

        :param min_value: the minimum permitted value
        :param max_value: the maximum permitted value
        """
        #: the lower bound, i.e., the minimum permitted value
        self.min_value: Final = min_value
        #: the upper bound, i.e., the maximum permitted value
        self.max_value: Final = max_value

    def __str__(self) -> str:
        """
        Get the name of these bounds.

        :return: the bounds separated by underscore

        >>> print(Bounds(-1, 1))
        m1_1
        >>> print(Bounds(0.5, 10.0))
        0d5_10
        >>> print(Bounds(-0.5, None))
        m0d5_n
        >>> print(Bounds(None, None))
        n_n
        """
        mi = "n" if self.min_value is None \
            else num_to_str_for_name(self.min_value)
        ma = "n" if self.max_value is None \
            else num_to_str_for_name(self.max_value)
        return f"{mi}_{ma}"

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        """
        Log the parameters of this bounds object to the given logger.

        :param logger: the logger for the parameters

        >>> from moptipy.utils.logger import InMemoryLogger
        >>> with InMemoryLogger() as l:
        ...     with l.key_values("C") as kv:
        ...         Bounds(1, 2).log_parameters_to(kv)
        ...     print(l.get_log())
        ['BEGIN_C', 'min: 1', 'max: 2', 'END_C']
        >>> with InMemoryLogger() as l:
        ...     with l.key_values("C") as kv:
        ...         Bounds(1.0, None).log_parameters_to(kv)
        ...     print(l.get_log())
        ['BEGIN_C', 'min: 1', 'min(hex): 0x1.0000000000000p+0', 'END_C']
        """
        if self.min_value is not None:
            logger.key_value(KEY_MIN, self.min_value,
                             also_hex=isinstance(self.min_value, float))
        if self.max_value is not None:
            logger.key_value(KEY_MAX, self.max_value,
                             also_hex=isinstance(self.max_value, float))


class FloatBounds(Bounds):
    """A class representing floating point number bounds."""

    def __init__(self, min_value: float = -1.0,
                 max_value: float = 1.0) -> None:
        """
        Initialize the floating point bounds.

        :param min_value: the minimum value
        :param max_value: the maximum value
        """
        if not isinstance(min_value, float):
            raise type_error(min_value, "min_value", float)
        if not isfinite(min_value):
            raise ValueError(
                f"min_value must be finite, but is {min_value}.")
        if not isinstance(max_value, float):
            raise type_error(max_value, "max_value", float)
        if not isfinite(max_value):
            raise ValueError(
                f"max_value must be finite, but is {max_value}.")
        if min_value >= max_value:
            raise ValueError(
                f"max_value > min_value must hold, but got "
                f"min_value={min_value} and max_value={max_value}.")
        super().__init__(min_value, max_value)


class OptionalFloatBounds(Bounds):
    """A class representing optional floating point number bounds."""

    def __init__(self, min_value: Optional[float] = None,
                 max_value: Optional[float] = None) -> None:
        """
        Initialize the optional floating point bounds.

        :param min_value: the optional minimum value
        :param max_value: the optional maximum value
        """
        if min_value is not None:
            if not isinstance(min_value, float):
                raise type_error(min_value, "min_value", float)
            if not isfinite(min_value):
                if min_value <= -inf:
                    min_value = None
                else:
                    raise ValueError(
                        f"min_value must be finite, but is {min_value}.")
        if max_value is not None:
            if not isinstance(max_value, float):
                raise type_error(max_value, "max_value", float)
            if not isfinite(max_value):
                if max_value >= inf:
                    max_value = None
                else:
                    raise ValueError(
                        f"max_value must be finite, but is {max_value}.")
        if (min_value is not None) and (max_value is not None) and \
                (min_value >= max_value):
            raise ValueError(
                f"max_value > min_value must hold, but got "
                f"min_value={min_value} and max_value={max_value}.")
        super().__init__(min_value, max_value)


class IntBounds(Bounds):
    """A class representing integer bounds."""

    def __init__(self, min_value: int = 0, max_value: int = 1) -> None:
        """
        Initialize the integer bounds.

        :param min_value: the minimum value
        :param max_value: the maximum value
        """
        if not isinstance(min_value, int):
            raise type_error(min_value, "min_value", int)
        if not isinstance(max_value, int):
            raise type_error(max_value, "max_value", int)
        if min_value >= max_value:
            raise ValueError(
                f"max_value > min_value must hold, but got "
                f"min_value={min_value} and max_value={max_value}.")
        super().__init__(min_value, max_value)


class OptionalIntBounds(Bounds):
    """A class representing optional integer bounds."""

    def __init__(self, min_value: Optional[int] = None,
                 max_value: Optional[int] = None) -> None:
        """
        Initialize the optional integer bounds.

        :param min_value: the optional minimum value
        :param max_value: the optional maximum value
        """
        if min_value is not None:
            if not isinstance(min_value, int):
                raise type_error(min_value, "min_value", int)
        if max_value is not None:
            if not isinstance(max_value, int):
                raise type_error(max_value, "max_value", int)
            if (min_value is not None) and (min_value >= max_value):
                raise ValueError(
                    f"max_value > min_value must hold, but got "
                    f"min_value={min_value} and max_value={max_value}.")
        super().__init__(min_value, max_value)



"""Test stuff on real vectors."""

from typing import Callable, Union, Iterable, List, Final, Tuple

from numpy.random import default_rng, Generator

from moptipy.api.algorithm import Algorithm
from moptipy.api.objective import Objective
from moptipy.examples.vectors.ackley import Ackley
from moptipy.spaces.bounded_vectorspace import BoundedVectorSpace
from moptipy.spaces.vectorspace import VectorSpace
from moptipy.tests.algorithm import validate_algorithm
from moptipy.utils.types import type_error

#: The dimensions for tests
DIMENSIONS_FOR_TESTS: Final[Tuple[int, ...]] = (1, 2, 3, 4, 5, 10)


def vectors_for_tests() -> Iterable[VectorSpace]:
    """
    Get a sequence of vector spaces for tests.

    :returns: the sequence of vector spaces
    """
    random: Final[Generator] = default_rng()
    spaces: Final[List[VectorSpace]] = []
    for dim in DIMENSIONS_FOR_TESTS:
        if random.integers(2) <= 0:
            spaces.append(VectorSpace(dim))
        else:
            a = b = float(random.uniform(-100.0, 100.0))
            while abs(a - b) < 1e-3:
                b = float(random.uniform(-100.0, 100.0))
            if a > b:
                a, b = b, a
            spaces.append(BoundedVectorSpace(dim, a, b))
    return spaces


def validate_algorithm_on_vectors(
        objective: Union[Objective, Callable[[VectorSpace], Objective]],
        algorithm: Union[Algorithm,
                         Callable[[VectorSpace, Objective], Algorithm]],
        max_fes: int = 100,
        uses_all_fes_if_goal_not_reached=True) -> None:
    """
    Check the validity of a black-box algorithm on vector problems.

    :param algorithm: the algorithm or algorithm factory
    :param objective: the objective function or function factory
    :param max_fes: the maximum number of FEs
    :param uses_all_fes_if_goal_not_reached: will the algorithm use all FEs
        unless it reaches the goal?
    """
    if not (isinstance(algorithm, Algorithm) or callable(algorithm)):
        raise type_error(algorithm, 'algorithm', Algorithm, True)
    if not (isinstance(objective, Objective) or callable(objective)):
        raise type_error(objective, "objective", Objective, True)

    for space in vectors_for_tests():
        if callable(objective):
            objf = objective(space)
            if not isinstance(objf, Objective):
                raise type_error(objf, "result of callable objective",
                                 Objective)
        else:
            objf = objective
        if callable(algorithm):
            algo = algorithm(space, objf)
            if not isinstance(algo, Algorithm):
                raise type_error(algo, "result of callable algorithm",
                                 Algorithm)
        else:
            algo = algorithm

        validate_algorithm(
            algorithm=algo, solution_space=space, objective=objf,
            max_fes=max_fes,
            uses_all_fes_if_goal_not_reached=uses_all_fes_if_goal_not_reached)


def validate_algorithm_on_ackley(
        algorithm: Union[Algorithm, Callable[
            [VectorSpace, Objective], Algorithm]],
        uses_all_fes_if_goal_not_reached: bool = True) -> None:
    """
    Check the validity of a black-box algorithm on Ackley's function.

    :param algorithm: the algorithm or algorithm factory
    :param uses_all_fes_if_goal_not_reached: will the algorithm use all FEs
        unless it reaches the goal?
    """
    validate_algorithm_on_vectors(
        Ackley(), algorithm,
        uses_all_fes_if_goal_not_reached=uses_all_fes_if_goal_not_reached)



"""An implementation of an unconstrained n-dimensional continuous space."""

from typing import Final

import numpy

from moptipy.spaces.nparrayspace import NPArraySpace
from moptipy.utils.nputils import is_np_float, is_all_finite

#: The default numerical datatype
_DEFAULT_TYPE: Final[numpy.dtype] = numpy.dtype(numpy.float64)


class VectorSpace(NPArraySpace):
    """
    A vector space where each element is a one-dimensional numpy float array.

    Such spaces are useful for continuous optimization.
    """

    def __init__(self, dimension: int,
                 dtype: numpy.dtype = _DEFAULT_TYPE) -> None:
        """
        Create the vector-based search space.

        :param dimension: The dimension of the search space,
            i.e., the number of decision variables.
        :param dtype: The basic data type of the vector space,
            i.e., the type of the decision variables
        """
        super().__init__(dimension, dtype)
        if not is_np_float(self.dtype):
            raise TypeError(f"Invalid data type {dtype}.")

    def validate(self, x: numpy.ndarray) -> None:
        """
        Validate a vector.

        :param x: the real vector
        :raises TypeError: if the vector is not an element of this space.
        :raises ValueError: if the shape of the vector is wrong or any of its
            element is not finite.
        """
        super().validate(x)
        if not is_all_finite(x):
            raise ValueError("All elements must be finite.")

    def n_points(self) -> int:
        """
        Get the number of different floating point values in this space.

        :return: The space contains unrestricted floating point numbers, so we
            return the approximate number of finite floating point numbers.

        >>> print(VectorSpace(3, numpy.dtype(numpy.float64)).n_points())
        6267911251143764491534102180507836301813760039183993274367
        """
        if self.dtype.char == "e":
            exponent = 5
            mantissa = 10
        elif self.dtype.char == "f":
            exponent = 8
            mantissa = 23
        elif self.dtype == "d":
            exponent = 11
            mantissa = 52
        elif self.dtype == "g":
            exponent = 15
            mantissa = 112
        else:
            raise ValueError(f"Invalid dtype {self.dtype}.")

        base = 2 * ((2 ** exponent) - 1) * (2 ** mantissa) - 1
        return base ** self.dimension

    def __str__(self) -> str:
        """
        Get the name of this space.

        :return: "vector" + dimension + dtype.char

        >>> print(VectorSpace(3, numpy.dtype(numpy.float64)))
        vector3d
        """
        return f"vector{self.dimension}{self.dtype.char}"
